/*
 * Model 3 Battery Program.  Uses EVTV ESP32 CANDue device to monitor MOdel 3 Vehicle CAN bus
 * and derive High Voltage Battery assembly data from CAN messages from the battery
 * 
 * Jack Rickard. 2019
 */
 
#include "esp32_can.h"        // CAN library
#include "battery.h"
#include "BufferedSerial.h"

template<class T> inline Print &operator <<(Print &obj, T arg) {
  obj.print(arg);
  return obj;
}

BATTERY battery;
CAN_FRAME inFrame1;

float version = 1.02;
bool debug = false;
bool debug2 = false;
uint16_t screenseconds=0;

char cmdBuffer[80];  //Buffer and pointer used for keyboard input over USB
int ptrBuffer = 0;
char globalstringvariable[80]; //Sample variables for command input
int globalintegervariable=0;
float globalfloatvariable=0;
String BatterySerialNumber="";

//************** PROGRAM INITIALIZATION ***********************
//This function will be performed once on bootup

void setup() {
  
  Serial.begin(115200);
  delay(7000);
  Serial.println("Initializing CAN....");
  initializeCAN();
  battery.begin(0);
}

//************** END PROGRAM INITIALIZATION ***********************


//************** MAIN PROGRAM LOOP ***********************

//This function will be repeated continuously during program operation

void loop() 
{
   if(screenseconds!=seconds()){screenseconds=seconds();printMain();}   

   checkforCAN0();    //Check for incoming CAN messages

   serialEvent();  //Check for USB keyboard input
}

//************** END MAIN PROGRAM LOOP ***********************



//************** PRINT SCREEN FUNCTIONS ***********************

void printMain()
{
  if(battery.SerialNumber.length()>8)BatterySerialNumber=battery.SerialNumber;
  buffFlush();
  buffPrintf("\f\n\n==========================================================================================\n"); 
  buffPrintf("******* EVTV Model 3 Battery Monitor Version ESP32-%2.3f ", version); 
  buffPrintf("Runtime: %02d Days %02d:%02d:%02d ******* ", days(), hours(), minutes(), seconds());
  buffPrintf("\n==========================================================================================\n");
  for (int i = 1; i <= 96; i=i+8) 
    {
      buffPrintf("  Cell %02d: %3.3f  %02d: %3.3f  %02d: %3.3f  %02d: %3.3f  %02d: %3.3f  %02d: %3.3f  %02d: %3.3f  %02d:%3.3f \n",i,battery.cell[i], i+1,battery.cell[i+1],i+2,battery.cell[i+2],i+3,battery.cell[i+3],i+4,battery.cell[i+4],i+5,battery.cell[i+5],i+6,battery.cell[i+6],i+7,battery.cell[i+7]);
    }

   buffPrintf("\n  Summed Voltage:%3.3fv Average Cell Voltage:%3.3fv \n", battery.Voltage, battery.AverageCell);
// buffPrintf("\n  Reported Voltage:%3.3fv Smooth Current:%3.3fA   Raw Current:%3.3fA Power:%3.0f watts\n", battery.reportedVoltage, battery.Amperes, battery.rawCurrent,battery.reportedVoltage * battery.Amperes);
   buffPrintf("\n  Reported Voltage:%3.3fv Current:%3.3fA  Power:%3.0f watts\n", battery.reportedVoltage, battery.Amperes, battery.reportedVoltage * battery.Amperes);
// buffPrintf("\n  SOC:%3.2f%%  minSoc:%3.2f%% maxSoc:%3.2f%% avgSoc:%3.2f%%   \n", battery.SOC,battery.minSOC,battery.maxSOC, battery.avgSOC );
   buffPrintf("\n  Capacity:%3.3f kWh  Currently:%3.3f kWh  Needed:%3.3f kWh  SOC:%3.2f%%  \n", battery.kWCapacity, battery.kWRemaining, battery.kWNeeded,battery.SOC);
   buffPrintf("\n  Lifetime Charge:%3.3f kWh  Lifetime Discharge:%3.3f kWh  \n", battery.lifeCharge, battery.lifeDischarge);
   Serial<<"  Battery Serial Number: "<<BatterySerialNumber<<"\n";
 
}
//************** END PRINT SCREEN FUNCTIONS ***********************



//************** CAN FUNCTIONS ***********************

void initializeCAN()
{
  CAN0.enable();
  if (CAN0.begin(500000,255))
    {
      CAN0.watchFor();
      Serial.println("Using CAN0, ESP32 native CAN port - initialization completed.\n");
    }
  else Serial << "CAN0 initialization (sync) ERROR\n";
}


void checkforCAN0()
{
   if (CAN0.available())
    {
      CAN0.read(inFrame1); 
      switch (inFrame1.id)
        {
          case 0x292:
          case 0x132:
          case 0x352:
          case 0x3D2:
          case 0x401:
          case 0x72A:
            battery.gotFrame(&inFrame1);
            break;
       
          default:
            handleCANframe(&inFrame1);
            break;      
        }
       if(debug) printFrame(&inFrame1, 0);
      }
}




void printFrame(CAN_FRAME *frame, int sent)
{
  char buffer[400];
  sprintf(buffer, "msgID 0x%03X; %02X; %02X; %02X; %02X; %02X; %02X; %02X; %02X  %02d:%02d:%02d.%04d\n", frame->id, frame->data.bytes[0],
          frame->data.bytes[1], frame->data.bytes[2], frame->data.bytes[3], frame->data.bytes[4], frame->data.bytes[5], frame->data.bytes[6],
          frame->data.bytes[7], hours(), minutes(), seconds(), milliseconds());

  if (sent)Serial << "Sent ";
  else Serial << "Received ";
  Serial << buffer << "\n";
}



void handleCANframe(CAN_FRAME *frame)
//This routine handles CAN interrupts from a capture of any other CAN frames from the inverter not specifically targeted.
//If you add other specific frames, do a setRXFilter and CallBack mailbox for each and do a method to catch those interrupts after the fashion
//of this one.
{
  //This routine basically just prints the general frame received IF debug is set.  Beyond that it does nothing.

  if (debug) printFrame(frame, 0); //If DEBUG variable is 1, print the actual message frame with a time stamp showing the time received.

}

//**************  END CAN FUNCTIONS ***********************


//**************  TIME FUNCTIONS ***********************

int milliseconds(void)
{
  int milliseconds = (int) (micros() / 100) % 10000 ;
  return milliseconds;
}

int secondtenths(void)
{
  int seconds = (int) (millis() / 1000) % 3 ;
  return seconds;
}

int seconds(void)
{
  int seconds = (int) (millis() / 1000) % 60 ;
  return seconds;
}

int minutes(void)
{
  int minutes = (int) ((millis() / (1000 * 60)) % 60);
  return minutes;
}

int hours(void)
{
  int hours   = (int) ((millis() / (1000 * 60 * 60)) % 24);
  return hours;
}

int days(void)
{
  int days   = (int) ((millis() / (1000 * 60 * 60 * 24)) % 49);
  return days;
}

//**************  END TIME FUNCTIONS ***********************

//**************  KEYBOARD FUNCTIONS ***********************

void serialEvent() 
{
  int incoming;
  incoming = Serial.read();
  if (incoming == -1) return; //false alarm....

  if (incoming == 10 || incoming == 13) //command done. Parse it.
    {
      handleConsoleCmd();
      ptrBuffer = 0; //reset line counter once the line has been processed
    }
  else
    {
      cmdBuffer[ptrBuffer++] = (unsigned char) incoming;
      if (ptrBuffer > 79) ptrBuffer = 79;
    }
}


void handleConsoleCmd()
{
  if (ptrBuffer == 1) handleShortCmd(); //command is a single ascii character
  else handleConfigCmd(); //if cmd over 1 char then assume that it is a config line
}


void handleConfigCmd()
{
  int i;
  int newValue;
  float newFloat;
 
  if (ptrBuffer < 6) return; //4 digit command, =, value is at least 6 characters
  cmdBuffer[ptrBuffer] = 0; //make sure to null terminate
  String cmdString = String();
  String textString = String();
  unsigned char whichEntry = '0';
  i = 0;

  while (cmdBuffer[i] != '=' && i < ptrBuffer)
    {
      cmdString.concat(String(cmdBuffer[i++]));
    }
  i++; //skip the =
  if (i >= ptrBuffer)
    {
      Serial.println("Command needs a value..ie CAPACITY=220");
      return; //or, we could use this to display the parameter instead of setting
    }

  // strtol() is able to parse also hex values (e.g. a string "0xCAFE"), useful for enable/disable by device id
  newValue = strtol((char *) (cmdBuffer + i), NULL, 0);
  newFloat = strtof((char *) (cmdBuffer + i), NULL);

  cmdString.toUpperCase();

  if (cmdString.startsWith("NAME"))
    {
      strcpy(globalstringvariable, cmdBuffer + i);   
    }
    
 
  else  if (cmdString.startsWith("UDP"))
    {
      if (newValue >= 6499 && newValue <= 6600)
        {
          globalintegervariable = newValue;
        }
    else Serial.println("\nInvalid Port. Enter a value between 6500 and 6599");
    }

  else if (cmdString.startsWith("LOV"))
    {
      if (newFloat >= 0.0f && newFloat <= 6.0f)
        {
          globalfloatvariable = newFloat;
        }
      else
        {
          Serial.println("Invalid lower cell voltage limit. Please enter a value 0.0 to 6.0");
        }
   }

  else if (cmdString.startsWith("RESTART")) 
    {
      ESP.restart();
    }
}

void handleShortCmd()
{
  uint8_t val;
  switch (cmdBuffer[0]) 
    {   
    
    case 'C':
    case 'c':
      battery.Voltage = battery.AverageCell= battery.reportedVoltage=battery.Amperes=battery.rawCurrent=0;
      battery.SOC=battery.kWCapacity=battery.kWRemaining=battery.kWNeeded=0;
      battery.lifeCharge=battery.lifeDischarge=0;
      for(uint8_t i=0;i<110;i++)
        {
          battery.cell[i]=0;
        }
      break;

    case 'D':
      debug2 = !debug2;
      break;
    case 'd':
      debug = !debug;
      break;
    }
}

//**************  END KEYBOARD FUNCTIONS ***********************
